import re
from datetime import date, datetime, timedelta

from django.conf import settings
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models import Count, Max, Q, QuerySet
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.template.loader import get_template
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django_fsm import FSMField, transition
from pandas.tseries.offsets import MonthEnd, YearEnd
from wbcore.contrib.authentication.models.users import User
from wbcore.contrib.color.enums import WBColor
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.enums import RequestType
from wbcore.markdown.utils import custom_url_fetcher
from wbcore.metadata.configs.buttons import ActionButton, ButtonDefaultColor
from wbcore.models import WBModel
from wbcore.models.fields import YearField
from wbcore.utils.models import ComplexToStringMixin
from weasyprint import HTML

from .compliance_type import ComplianceDocumentMixin, ComplianceType, can_active_request
from .enums import IncidentSeverity


def can_draft_request(instance, user: "User") -> bool:
    if instance.is_instance:
        return False
    return user.has_perm("wbcompliance.administrate_compliance")


class ComplianceTaskGroup(WBModel):
    name = models.CharField(max_length=255, verbose_name=_("Name"))
    order = models.PositiveIntegerField(null=True, blank=True, verbose_name=_("Order"))

    class Meta:
        verbose_name = "Compliance Task Group"
        verbose_name_plural = "Compliance Task Groups"

    def __str__(self) -> str:
        return "{}".format(self.name)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:compliancetaskgroup"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcompliance:compliancetaskgrouprepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name}}"


class ComplianceTask(WBModel):
    class Occurrence(models.TextChoices):
        YEARLY = "YEARLY", "Yearly"
        QUARTERLY = "QUARTERLY", "Quarterly"
        MONTHLY = "MONTHLY", "Monthly"
        NEVER = "NEVER", "Never"

        @classmethod
        def get_color_map(cls) -> list:
            colors = [WBColor.GREEN_LIGHT.value, WBColor.YELLOW_LIGHT.value, WBColor.RED_LIGHT.value]
            return [choice for choice in zip(cls, colors, strict=False)]

    title = models.CharField(max_length=255, verbose_name=_("Title"))
    description = models.TextField(default="", blank=True, verbose_name=_("Description"))
    occurrence = models.CharField(
        max_length=32,
        default=Occurrence.MONTHLY,
        choices=Occurrence.choices,
        verbose_name=_("Occurrence"),
    )
    active = models.BooleanField(default=True)
    group = models.ForeignKey(
        to="wbcompliance.ComplianceTaskGroup",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="tasks_related",
        verbose_name=_("Group"),
    )
    review = models.ManyToManyField(
        "wbcompliance.ReviewComplianceTask",
        related_name="tasks",
        blank=True,
        verbose_name=_("Review"),
        help_text=_("list of reviews that contain this task"),
    )
    risk_level = models.CharField(
        max_length=32,
        blank=True,
        null=True,
        choices=IncidentSeverity.choices,
        verbose_name=_("Risk Level"),
    )
    remarks = models.TextField(null=True, blank=True, verbose_name=_("Remarks"))
    type = models.ForeignKey(
        to=ComplianceType, on_delete=models.PROTECT, related_name="tasks_of_type", verbose_name=_("Type")
    )

    class Meta:
        verbose_name = "Compliance Task"
        verbose_name_plural = "Compliance Tasks"

    def generate_compliance_task_instance(self, occured: date | None = None, link_instance_review: bool = False):
        kwargs = {"task": self}
        if _instance := ComplianceTaskInstance.objects.filter(task=self).last():
            kwargs.update({"status": _instance.status, "text": _instance.text, "summary_text": _instance.summary_text})
        new_instance = ComplianceTaskInstance.objects.create(**kwargs)

        if occured:
            new_instance.occured = occured
            new_instance.save()

        if link_instance_review:
            for review in self.review.filter(Q(is_instance=False) & Q(review_task=None)):
                qs = ReviewComplianceTask.objects.filter(review_task=review, is_instance=True)
                if review_instance := qs.filter(occured=new_instance.occured).last():
                    new_instance.review.add(review_instance)
        return new_instance

    def __str__(self):
        return "{}".format(self.title)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:compliancetask"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcompliance:compliancetaskrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{title}}"


class ComplianceTaskInstance(models.Model):
    class Status(models.TextChoices):
        NOT_CHECKED = "NOT_CHECKED", "Not Checked"
        WARNING = "WARNING", "Warning"
        FOR_INFO = "FOR_INFO", "For Info"
        NOTHING_TO_REPORT = "NOTHING_TO_REPORT", "Nothing to Report"
        BREACH = "BREACH", "Breach"

        @classmethod
        def get_color_map(cls) -> list:
            colors = [
                WBColor.GREY.value,
                WBColor.YELLOW_LIGHT.value,
                WBColor.YELLOW.value,
                WBColor.BLUE_LIGHT.value,
                WBColor.RED_LIGHT.value,
            ]
            return [choice for choice in zip(cls, colors, strict=False)]

    task = models.ForeignKey(
        on_delete=models.CASCADE,
        to="wbcompliance.ComplianceTask",
        related_name="task_instances_related",
        verbose_name=_("Compliance Task"),
    )
    occured = models.DateField(auto_now_add=True, verbose_name=_("Occured"))
    status = models.CharField(
        max_length=32,
        default=Status.NOT_CHECKED,
        choices=Status.choices,
        verbose_name=_("Status"),
    )
    text = models.TextField(default="", blank=True, verbose_name=_("Text"))
    summary_text = models.TextField(null=True, blank=True, verbose_name=_("Summary Text"))
    review = models.ManyToManyField(
        "wbcompliance.ReviewComplianceTask",
        related_name="task_instances",
        blank=True,
        verbose_name=_("Review"),
        help_text=_("list of reviews that contain this task instance"),
    )

    class Meta:
        verbose_name = "Compliance Task Instance"
        verbose_name_plural = "Compliance Task Instances"

    @classmethod
    def get_max_depth(cls) -> int | None:
        return cls.objects.all().values("task").annotate(dcount=Count("task")).aggregate(max=Max("dcount")).get("max")

    @classmethod
    def get_dict_max_count_task(cls) -> dict:
        if cls.get_max_depth():
            return cls.objects.all().values("task").annotate(dcount=Count("task")).latest("dcount")
        return {}

    def __str__(self) -> str:
        return "{}".format(self.task.title)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:compliancetaskinstance"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{task__title}} : {{id}}"


class ComplianceAction(models.Model):
    class Status(models.TextChoices):
        TO_BE_DONE = "TO_BE_DONE", "To be done"
        WORK_IN_PROGRESS = "WORK_IN_PROGRESS", "Work in Progress"
        DONE = "DONE", "Done"

        @classmethod
        def get_color_map(cls) -> list:
            colors = [WBColor.GREY.value, WBColor.YELLOW_LIGHT.value, WBColor.GREEN_LIGHT.value]
            return [choice for choice in zip(cls, colors, strict=False)]

    title = models.CharField(max_length=255, verbose_name=_("Title"))
    description = models.TextField(
        default="",
        null=True,
        blank=True,
        verbose_name=_("Description"),
    )
    summary_description = models.TextField(
        null=True,
        blank=True,
        verbose_name=_("Summary Description"),
    )
    deadline = models.DateField(null=True, blank=True, verbose_name=_("Deadline"))
    progress = models.FloatField(
        default=0, validators=[MinValueValidator(0.0), MaxValueValidator(1.0)], verbose_name=_("Progress")
    )
    status = models.CharField(
        max_length=32,
        default=Status.TO_BE_DONE,
        choices=Status.choices,
        verbose_name=_("Status"),
    )
    type = models.ForeignKey(
        to=ComplianceType, on_delete=models.PROTECT, related_name="actions_of_type", verbose_name=_("Type")
    )
    active = models.BooleanField(default=True, verbose_name=_("Active"))
    creator = models.ForeignKey(
        to="directory.Person",
        null=True,
        blank=True,
        related_name="compliance_actions",
        verbose_name=_("Creator"),
        on_delete=models.SET_NULL,
    )
    created = models.DateTimeField(auto_now_add=True, verbose_name=_("Created"))
    changer = models.ForeignKey(
        to="directory.Person",
        null=True,
        blank=True,
        verbose_name=_("Changer"),
        related_name="updated_actions",
        on_delete=models.SET_NULL,
    )
    last_modified = models.DateTimeField(auto_now=True, verbose_name=_("Last modified"))

    class Meta:
        verbose_name = "Compliance Action"
        verbose_name_plural = "Compliance Actions"

    def __str__(self) -> str:
        return "{}".format(self.title)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:complianceaction"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{title}}"


class ComplianceEvent(models.Model):
    class Type(models.TextChoices):
        INCIDENT = "INCIDENT", "Incident"
        INFO = "INFO", "Info"

        @classmethod
        def get_color_map(cls) -> list:
            colors = [WBColor.GREY.value, WBColor.BLUE_LIGHT.value]
            return [choice for choice in zip(cls, colors, strict=False)]

    type_event = models.CharField(
        max_length=32,
        default=Type.INCIDENT,
        choices=Type.choices,
        verbose_name=_("Type Event"),
    )
    level = models.CharField(
        max_length=32,
        default=IncidentSeverity.LOW,
        choices=IncidentSeverity.choices,
        verbose_name=_("Level"),
    )
    title = models.CharField(max_length=255, verbose_name=_("Title"))
    exec_summary = models.TextField(
        null=True,
        blank=True,
        verbose_name=_("Executive Summary"),
    )
    exec_summary_board = models.TextField(
        null=True,
        blank=True,
        verbose_name=_("Executive Summary for the Board"),
    )
    description = models.TextField(default="", blank=True, verbose_name=_("Description"))
    actions_taken = models.TextField(default="", blank=True, verbose_name=_("Actions Taken"))
    consequences = models.TextField(default="", blank=True, verbose_name=_("Consequences"))
    future_suggestions = models.TextField(default="", blank=True, verbose_name=_("Future Suggestions"))
    type = models.ForeignKey(
        to=ComplianceType, on_delete=models.PROTECT, related_name="events_of_type", verbose_name=_("Type")
    )
    active = models.BooleanField(default=True)
    creator = models.ForeignKey(
        to="directory.Person",
        null=True,
        blank=True,
        related_name="compliance_events",
        verbose_name=_("Creator"),
        on_delete=models.SET_NULL,
    )
    created = models.DateTimeField(auto_now_add=True, verbose_name=_("Created"))
    changer = models.ForeignKey(
        to="directory.Person",
        null=True,
        blank=True,
        related_name="updated_events",
        verbose_name=_("Changer"),
        on_delete=models.SET_NULL,
    )
    last_modified = models.DateTimeField(auto_now=True, verbose_name=_("Last modified"))
    confidential = models.BooleanField(default=False, verbose_name=_("Confidential"))

    class Meta:
        verbose_name = "Compliance Event"
        verbose_name_plural = "Compliance Events"

        notification_types = [
            create_notification_type(
                code="wbcompliance.complianceevent.notify",
                title="Compliance Event Notification",
                help_text="Sends out a notification when a new compliance event was created.",
                email=True,
                web=False,
                mobile=False,
                is_lock=True,
            )
        ]

    def __str__(self) -> str:
        return "{}".format(self.title)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:complianceevent"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{title}}"


class ReviewComplianceTask(ComplianceDocumentMixin, ComplexToStringMixin, WBModel):
    class Occurrence(models.TextChoices):
        YEARLY = "YEARLY", "Yearly"
        QUARTERLY = "QUARTERLY", "Quarterly"
        MONTHLY = "MONTHLY", "Monthly"
        NEVER = "NEVER", "Never"

    class Status(models.TextChoices):
        DRAFT = "DRAFT", "Draft"
        VALIDATION_REQUESTED = "VALIDATION_REQUESTED", "Validation Requested"
        VALIDATED = "VALIDATED", "Validated"

        @classmethod
        def get_color_map(cls) -> list:
            colors = [
                WBColor.BLUE_LIGHT.value,
                WBColor.YELLOW_LIGHT.value,
                WBColor.GREEN_LIGHT.value,
            ]
            return [choice for choice in zip(cls, colors, strict=False)]

    class Meta:
        verbose_name = "Review Compliance Task"
        verbose_name_plural = "Review Compliance Tasks"

        notification_types = [
            create_notification_type(
                code="wbcompliance.reviewcompliancetask.notify",
                title="Compliance Task Review Notification",
                help_text="Notifies you when a compliance task can be reviewed",
                email=True,
                web=False,
                mobile=False,
                is_lock=True,
            )
        ]

    title = models.CharField(max_length=255, verbose_name=_("Title"))
    from_date = models.DateField(null=True, blank=True, verbose_name=_("From"))
    to_date = models.DateField(null=True, blank=True, verbose_name=_("To"))
    description = models.TextField(default="", blank=True, verbose_name=_("Description"))
    year = YearField(
        validators=[MinValueValidator(1000), MaxValueValidator(9999)], null=True, blank=True, verbose_name=_("Year")
    )
    creator = models.ForeignKey(
        to="directory.Person",
        null=True,
        blank=True,
        verbose_name=_("Creator"),
        related_name="author_review_compliane_tasks",
        on_delete=models.CASCADE,
    )
    created = models.DateTimeField(auto_now_add=True, verbose_name=_("Created"))
    changer = models.ForeignKey(
        "directory.Person", null=True, blank=True, verbose_name=_("Changer"), on_delete=models.deletion.SET_NULL
    )
    changed = models.DateTimeField(auto_now=True, verbose_name=_("Changed"))
    status = FSMField(default=Status.DRAFT, choices=Status.choices, verbose_name=_("Status"))
    occurrence = models.CharField(
        max_length=32,
        default=Occurrence.MONTHLY,
        choices=Occurrence.choices,
        verbose_name=_("Occurrence"),
    )

    is_instance = models.BooleanField(default=False, verbose_name=_("Is occurrence"))
    review_task = models.ForeignKey(
        to="wbcompliance.ReviewComplianceTask",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="review_tasks",
        verbose_name=_("Parent Review"),
    )
    occured = models.DateField(null=True, blank=True, verbose_name=_("Occured Instance"))
    type = models.ForeignKey(
        to=ComplianceType,
        on_delete=models.PROTECT,
        related_name="reviewtask_of_type",
        verbose_name=_("Type"),
    )

    def notify(self, title, msg, recipients: QuerySet["User"]) -> None:
        for user in recipients:
            send_notification(
                code="wbcompliance.reviewcompliancetask.notify",
                title=title,
                body=msg,
                user=user,
                reverse_name="wbcompliance:reviewcompliancetask-detail",
                reverse_args=[self.id],
            )

    def _remove_styled_paragraph(self, html_content: str) -> str:
        tags = ["p", "span"]
        for tag in tags:
            reg_str = "<" + tag + "(.*?)" + ">"
            reg_style = '(style=".*?;")'
            _attributs = re.findall(reg_str, html_content)
            dict_attrs = {}
            for attribut in _attributs:
                dict_attrs[attribut] = re.sub(reg_style, "", attribut)

            for _key, _value in dict_attrs.items():
                html_content = re.sub(_key, _value, html_content)

        html_content = re.sub("<p>&nbsp;</p>", "", html_content)
        return html_content

    def generate_pdf(self) -> bytes:
        html = get_template("compliance/review_compliance_task_report.html")
        table = {}
        tasks = ComplianceTask.objects.filter(Q(review=self) & Q(active=True)).order_by("group__order")
        group_ids = tasks.values_list("group", flat=True).distinct()
        for group_id in group_ids:
            if group_id:
                group = ComplianceTaskGroup.objects.get(id=group_id)
                table[group.id] = {"name": group.name, "tasks": {}}
            else:
                group_id = ""
                group = None
                table[""] = {"name": "", "tasks": {}}

            for task in tasks.filter(group=group):
                table[group_id]["tasks"][task.id] = {
                    "title": task.title,
                    "description": self._remove_styled_paragraph(task.description),
                    "risk_level": task.risk_level,
                    "remarks": task.remarks,
                }
        html_content = html.render(
            {
                "today": timezone.now(),
                "review": self,
                "table": table,
            }
        )
        return HTML(
            string=html_content, base_url=settings.BASE_ENDPOINT_URL, url_fetcher=custom_url_fetcher
        ).write_pdf()

    def get_period_date(self, today: datetime | None = None):
        if today is None:
            today = timezone.now()
        if self.occurrence == self.Occurrence.YEARLY:
            from_date = date(today.year - 1, 1, 1)
            to_date = from_date + YearEnd(1)
        elif self.occurrence == self.Occurrence.QUARTERLY:
            from_date = (today - MonthEnd(4)).date() + timedelta(days=1)
            to_date = (today - MonthEnd(1)).date()
        elif self.occurrence == self.Occurrence.MONTHLY:
            from_date = (today - MonthEnd(2)).date() + timedelta(days=1)
            to_date = (today - MonthEnd(1)).date()
        else:
            return None, None
        return from_date, to_date

    def generate_review_compliance_task_instance(
        self, current_date: datetime | None = None, link_instance: bool = False, notify_admin: bool = False
    ) -> None:
        """
        allow to generate the occurrence of the Indicators Report.
        current_date: allow to find the year and the period date corresponding to the occurrence. we use today by default
        """
        if current_date is None:
            current_date = timezone.now()
        from_date, to_date = self.get_period_date(current_date)
        date_title = to_date if to_date else current_date
        kwargs = {
            "review_task": self,
            "year": current_date.year,
            "occured": current_date.date(),
            "is_instance": True,
            "status": ReviewComplianceTask.Status.DRAFT,
            "creator": self.creator,
            "changer": self.changer,
            "occurrence": ReviewComplianceTask.Occurrence.NEVER,
            "description": self.description,
            "title": "{} - {}".format(self.title, date_title.strftime("%b %Y")),
            "from_date": from_date,
            "to_date": to_date,
            "type": self.type,
        }
        if _instance := ReviewComplianceTask.objects.filter(Q(review_task=self) & Q(is_instance=True)).last():
            kwargs.update({"creator": _instance.changer, "description": _instance.description})
        new_review = ReviewComplianceTask.objects.create(**kwargs)

        if link_instance:
            for task in ComplianceTask.objects.filter(review__in=[self]):
                if instance := ComplianceTaskInstance.objects.filter(task=task).last():
                    instance.review.add(new_review)

        if notify_admin and (compliance_type := self.type):
            recipients = ComplianceType.get_administrators(compliance_type)
            msg = _("<b>{}</b> is available. You can now complete and validate it.").format(self.title)
            title = _("Instance Indicator Report: {}").format(self.title)
            self.notify(title, msg, recipients)

    def get_task_group_ids_from_review(self, through_task: bool = True, task_with_group: bool = True) -> set:
        if through_task:
            return set(
                self.tasks.filter(group__isnull=not (task_with_group))
                .order_by("group__order")
                .values_list("group", flat=True)
            )
        else:
            return set(
                self.task_instances.filter(task__group__isnull=not (task_with_group))
                .order_by("task__group__order")
                .values_list("task__group", flat=True)
            )

    @transition(
        field=status,
        source=Status.DRAFT,
        target=Status.VALIDATION_REQUESTED,
        permission=lambda _, user: user.has_perm("wbcompliance.administrate_compliance"),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbcompliance:reviewcompliancetask",),
                icon=WBIcon.SEND.icon,
                key="validationrequested",
                label="Request Validation",
                action_label=_("Request Validation"),
                description_fields=_(
                    "<p>Title: <b>{{title}}</b></p>\
                <p>Status: <b>{{status}}</b></p> <p>From: <b>{{from_date}}</b></p>\
                <p>To: <b>{{to_date}}</b></p> <p>Do you want to send this request for validation ?</p>"
                ),
            )
        },
    )
    def validationrequested(self, by=None, description=None, **kwargs):
        if compliance_type := self.type:
            # notify the compliance team without the current user
            if by:
                self.changer = by.profile
            current_user = self.changer if self.changer else self.creator
            recipients = ComplianceType.get_administrators(compliance_type).exclude(profile=current_user)
            msg = _("Validation Request from {} to validate a ompliance Risk Review: <b>{}</b>").format(
                str(current_user), self.title
            )
            title = _("Validation Requested Compliance Risk Review: {}").format(self.title)
            self.notify(title, msg, recipients)

    @transition(
        field=status,
        source=Status.VALIDATION_REQUESTED,
        target=Status.DRAFT,
        permission=lambda _, user: user.has_perm("wbcompliance.administrate_compliance"),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbcompliance:reviewcompliancetask",),
                icon=WBIcon.EDIT.icon,
                key="draft",
                label="Return to Draft Mode",
                action_label=_("Return to Draft Mode"),
                description_fields=_(
                    """
                <p>Title: <b> {{title}} </b></p>
                <p>Status: <b>{{status}}</b></p> <p>From: <b>{{from_date}}</b></p>  <p>To: <b>{{to_date}}</b></p>
                <p>Do you want to return to draft ?</p>
                """
                ),
            )
        },
    )
    def draft(self, by=None, description=None, **kwargs):
        if compliance_type := self.type:
            if by:
                self.changer = by.profile
            current_user = self.changer if self.changer else self.creator
            msg = _("{} has changed a Compliance Risk Review to Draft : <b>{}</b>").format(
                str(current_user), self.title
            )
            title = _("Compliance Risk Review : {}").format(self.title)
            recipients = ComplianceType.get_administrators(compliance_type).exclude(profile=current_user)
            self.notify(title, msg, recipients)

    @transition(
        field=status,
        source=Status.VALIDATION_REQUESTED,
        target=Status.VALIDATED,
        permission=can_active_request,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbcompliance:reviewcompliancetask",),
                icon=WBIcon.SEND.icon,
                key="validation",
                label="Validate",
                action_label=_("Validate"),
                description_fields=_(
                    """
                <p>Title: <b> {{title}} </b></p>
                <p>Status: <b>{{status}}</b></p> <p>From: <b>{{from_date}}</b></p>  <p>To: <b>{{to_date}}</b></p>
                <p>Do you want to validate?</p>
                """
                ),
            )
        },
    )
    def validation(self, by=None, description=None, **kwargs):
        if compliance_type := self.type:
            if by:
                self.changer = by.profile
            current_user = self.changer if self.changer else self.creator
            msg = _(
                """
                {} has validated a Compliance Risk Review : <b>{}</b>
            """
            ).format(str(current_user), self.title)
            title = _("Validation - Compliance Risk Review : {}").format(self.title)
            recipients = ComplianceType.get_administrators(compliance_type).exclude(profile=current_user)
            self.notify(title, msg, recipients)

    @transition(
        field=status,
        source=Status.VALIDATED,
        target=Status.DRAFT,
        permission=can_draft_request,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbcompliance:reviewcompliancetask",),
                icon=WBIcon.EDIT.icon,
                key="backtodraft",
                label="Back to draft",
                action_label=_("Back to draft"),
                description_fields=_(
                    """
                <p>Title: <b> {{title}} </b></p>
                <p>Status: <b>{{status}}</b></p> <p>From: <b>{{from_date}}</b></p>  <p>To: <b>{{to_date}}</b></p>
                <p>Do you want to return to the draft?</p>
                """
                ),
            )
        },
    )
    def backtodraft(self, by=None, description=None, **kwargs):
        if compliance_type := self.type:
            if by:
                self.changer = by.profile
            current_user = self.changer if self.changer else self.creator
            msg = _("{} has drafted a Compliance Indicators Report : <b>{}</b>").format(str(current_user), self.title)
            title = _("Compliance Indicators Report drafted: {}").format(self.title)
            recipients = ComplianceType.get_administrators(compliance_type)
            self.notify(title, msg, recipients)

    def compute_str(self) -> str:
        _str = "{}".format(self.title)
        if self.from_date or self.to_date:
            _str += " - ({} - {})".format(self.from_date, self.to_date)
        return _str

    def __str__(self) -> str:
        return self.computed_str

    def save(self, *args, **kwargs):
        self.computed_str = self.compute_str()
        super().save(*args, **kwargs)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcompliance:reviewcompliancetask"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcompliance:reviewcompliancetaskrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{computed_str}}"


@receiver(post_save, sender=ComplianceEvent)
def post_save_compliance_event(sender, instance, created, **kwargs):
    """
    Send notification to administrators
    """
    if created and (compliance_type := instance.type):
        current_profile = instance.changer if instance.changer else instance.creator
        recipients = ComplianceType.get_administrators(compliance_type).exclude(profile=current_profile)
        title = "{}: {}".format(ComplianceEvent.Type[instance.type_event].label, instance.title)
        msg = _("<p>An {} Event was created by {} {} at {}</p>").format(
            ComplianceEvent.Type[instance.type_event].label,
            current_profile.first_name,
            current_profile.last_name,
            instance.last_modified.strftime("%d-%b-%y %H:%M:%S"),
        )
        if instance.exec_summary:
            msg += _("<p> Summary : {}</p>").format(instance.exec_summary)
        for recipient in recipients:
            send_notification(
                code="wbcompliance.complianceevent.notify",
                title=title,
                body=msg,
                user=recipient,
                reverse_name="wbcompliance:complianceevent-detail",
                reverse_args=[instance.id],
            )



from contextlib import suppress

from rest_framework.reverse import reverse
from wbcore.metadata.configs import buttons as bt
from wbcore.signals import add_instance_button

from wbcompliance.models.risk_management.checks import RiskCheck


class RiskCheckViewSetMixin:
    @classmethod
    def _get_risk_checks_button_title(cls) -> str:
        return "Checks"

    def _add_risk_check_button(self, sender, many, *args, view=None, **kwargs):
        with suppress(AssertionError):
            if view and (instance := view.get_object()):
                icon = RiskCheck.CheckStatus[instance.get_worst_check_status()].icon
                if instance.checks.exists():
                    return bt.WidgetButton(
                        endpoint=f'{reverse("wbcompliance:riskcheck-list", args=[], request=self.request)}?checked_objects=[[{instance.checked_object_content_type.id},{instance.checked_object.id}]]&evaluation_date={instance.check_evaluation_date}&activators=[[{instance.activator_content_type.id},{instance.activator_id}]]&passive_check=False',
                        label=self._get_risk_checks_button_title(),
                        icon=icon,
                    )

    def options(self, request, *args, **kwargs):
        """
        Handler method for HTTP 'OPTIONS' request.
        """
        add_instance_button.connect(
            self._add_risk_check_button,
            sender=self.__class__,
            dispatch_uid="wbcompliance_add_instance_button_riskcheck",
        )
        return super().options(request, *args, **kwargs)



from collections import defaultdict
from datetime import date, datetime

from celery import shared_task
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from tqdm import tqdm
from wbcore.workers import Queue

from wbcompliance.models import ComplianceForm, ComplianceFormSignature, ComplianceTask, ReviewComplianceTask
from wbcompliance.models.risk_management.rules import (
    RiskRule,
    RuleCheckedObjectRelationship,
)


@shared_task(queue=Queue.EXTENDED_BACKGROUND.value)
def check_passive_rules(
    from_date: date | None = None,
    to_date: date | None = None,
    override_incident: bool | None = False,
    extra_process_kwargs: dict | None = None,
    silent_notification: bool = False,
    debug: bool = False,
):
    """
    Periodic function that call all active passive rules and checked_object the check workflow
    """
    if not to_date:
        to_date = datetime.today().date()

    # cleanup relationship before continuing
    clean_dynamic_rule_relationships(debug=debug)

    process_kwargs = {
        "override_incident": override_incident,
    }
    if isinstance(extra_process_kwargs, dict):
        process_kwargs.update(extra_process_kwargs)

    res = []
    for rule in RiskRule.objects.filter(is_enable=True):
        for relationship in rule.checked_object_relationships.iterator():
            res.append((rule, relationship))
    gen = res
    if debug:
        gen = tqdm(gen, total=len(res))
    date_to_notify = defaultdict(set)
    for rule, relationship in gen:
        for evaluation_date in relationship.get_unchecked_dates(from_date=from_date, to_date=to_date):
            if relationship.process_rule(evaluation_date, **process_kwargs):
                date_to_notify[rule].add(evaluation_date)

    if not silent_notification:
        for rule, dates in date_to_notify.items():
            for checked_date in dates:
                rule.notify(checked_date)


@shared_task(queue=Queue.BACKGROUND.value)
def clean_dynamic_rule_relationships(debug: bool = False):
    """
    Periodic function to check reverse generic object that don't exist anymore. Furthermore, get the queryset representing all the active relationship for all backend and ensure that every object have a relationship (e.g. new object creation)
    """
    gen = RiskRule.objects.filter(apply_to_all_active_relationships=True)
    if debug:
        gen = tqdm(gen, total=gen.count())

    for rule in gen:
        leftover_relationships = rule.checked_object_relationships.all()
        for content_object in rule.rule_backend.get_all_active_relationships():
            rel, _ = RuleCheckedObjectRelationship.objects.get_or_create(
                rule=rule,
                checked_object_content_type=ContentType.objects.get_for_model(content_object),
                checked_object_id=content_object.id,
            )
            leftover_relationships = leftover_relationships.exclude(id=rel.id)

        for leftover_relationship in leftover_relationships.iterator():
            leftover_relationship.delete()


@shared_task(queue=Queue.BACKGROUND.value)
def periodic_quaterly_or_monthly_compliance_task():
    today = datetime.now()
    qs = ComplianceTask.objects.filter(active=True)
    qs_review = ReviewComplianceTask.objects.filter(
        Q(status=ReviewComplianceTask.Status.VALIDATED) & Q(is_instance=False)
    )
    if today.month == 1 and today.day == 1:
        qs = qs.filter(
            Q(occurrence=ComplianceTask.Occurrence.YEARLY) | Q(occurrence=ComplianceTask.Occurrence.MONTHLY)
        )
        qs_review = qs_review.filter(
            Q(occurrence=ReviewComplianceTask.Occurrence.YEARLY)
            | Q(occurrence=ReviewComplianceTask.Occurrence.MONTHLY)
        )
    elif today.month % 3 == 0:
        qs = qs.filter(
            Q(occurrence=ComplianceTask.Occurrence.QUARTERLY) | Q(occurrence=ComplianceTask.Occurrence.MONTHLY)
        )
        qs_review = qs_review.filter(
            Q(occurrence=ReviewComplianceTask.Occurrence.QUARTERLY)
            | Q(occurrence=ReviewComplianceTask.Occurrence.MONTHLY)
        )
    else:
        qs = qs.filter(occurrence=ComplianceTask.Occurrence.MONTHLY)
        qs_review = qs_review.filter(occurrence=ReviewComplianceTask.Occurrence.MONTHLY)

    for review_task in qs_review:
        review_task.generate_review_compliance_task_instance(notify_admin=True)

    for task in qs:
        task.generate_compliance_task_instance(link_instance_review=True)


@shared_task
def send_reminder_of_unsigned_compliance_form(compliance_form_ids: list[int] | None = None) -> None:
    """
    Send a reminder to the user to sign the compliance form.
    """

    compliance_forms = (
        ComplianceForm.objects.filter(id__in=compliance_form_ids)
        if compliance_form_ids
        else ComplianceForm.objects.filter(status=ComplianceForm.Status.ACTIVE)
    )

    for compliance_form in compliance_forms:
        signatures = ComplianceFormSignature.objects.filter(
            compliance_form=compliance_form, version=compliance_form.version, signed=None
        )
        recipients = get_user_model().objects.filter(profile__id__in=signatures.values("person"))
        if recipients.exists():
            msg = f"<p>{compliance_form.changer} has activated a {compliance_form.title} {compliance_form.form_type} version {compliance_form.version} </p>"
            if compliance_form.policy and compliance_form.policy != "<p></p>" and compliance_form.policy != "null":
                msg += f"</br><p><b>User's ComplianceForm:</b></p><i>{compliance_form.policy}</i>"
            title = f"Unsigned Compliance Form - {compliance_form.form_type} Reminder : {compliance_form.title}"
            compliance_form.notify(title, msg, recipients=recipients)



# Generated by Django 4.2.7 on 2023-11-06 13:30

import django.core.serializers.json
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "wbcompliance",
            "0001_initial_squashed_squashed_0010_alter_checkedobjectincidentrelationship_resolved_by_and_more",
        ),
    ]

    operations = [
        migrations.AlterField(
            model_name="riskrule",
            name="parameters",
            field=models.JSONField(blank=True, default=dict, encoder=django.core.serializers.json.DjangoJSONEncoder),
        ),
    ]



from .compliance_form import (
    ComplianceFormDisplayConfig,
    ComplianceFormRuleDisplayConfig,
    ComplianceFormSectionDisplayConfig,
    ComplianceFormSignatureDisplayConfig,
    ComplianceFormSignatureSectionRuleDisplayConfig,
    ComplianceFormTypeDisplayConfig,
)
from .compliance_task import (
    ComplianceActionDisplayConfig,
    ComplianceEventDisplayConfig,
    ComplianceTaskDisplayConfig,
    ComplianceTaskGroupDisplayConfig,
    ComplianceTaskInstanceComplianceTaskDisplayConfig,
    ComplianceTaskInstanceDisplayConfig,
    ComplianceTaskInstanceReviewDisplayConfig,
    ComplianceTaskMatrixPandasDisplayConfig,
    ComplianceTaskReviewDisplayConfig,
    ReviewComplianceTaskDisplayConfig,
)
from .compliance_type import ComplianceTypeDisplayConfig
from .risk_managment import *



from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.contrib.messages.api import warning
from django.contrib.postgres.fields.ranges import DateRangeField
from django.db.models import (
    Case,
    CharField,
    DateField,
    Exists,
    ExpressionWrapper,
    F,
    IntegerField,
    OuterRef,
    Subquery,
    Value,
    When,
)
from django.db.models.functions import Cast, Concat
from django.http import HttpResponse
from django.utils.functional import cached_property
from rest_framework import status
from rest_framework.decorators import action
from wbcore import viewsets
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.permission.models import UserObjectPermission

from wbcompliance.filters import (
    CheckedObjectIncidentRelationshipFilterSet,
    RiskIncidentFilterSet,
)
from wbcompliance.models import RuleThreshold
from wbcompliance.models.risk_management import (
    CheckedObjectIncidentRelationship,
    RiskIncident,
    RiskIncidentType,
    RiskRule,
)
from wbcompliance.models.risk_management.incidents import resolve_all_incidents_as_task
from wbcompliance.permissions import RulePermission
from wbcompliance.serializers import (
    CheckedObjectIncidentRelationshipModelSerializer,
    CheckedObjectIncidentRelationshipRepresentationSerializer,
    RiskIncidentModelSerializer,
    RiskIncidentRepresentationSerializer,
    RiskIncidentTypeRepresentationSerializer,
)
from wbcompliance.viewsets.buttons import RiskIncidentButtonConfig
from wbcompliance.viewsets.display import (
    CheckedObjectIncidentRelationshipRiskRuleDisplayConfig,
    RiskIncidentDisplayConfig,
)
from wbcompliance.viewsets.endpoints import (
    CheckedObjectIncidentRelationshipEndpointConfig,
    CheckedObjectIncidentRelationshipRiskRuleEndpointConfig,
    RiskIncidentEndpointConfig,
)


class RiskIncidentTypeRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = RiskIncidentType.objects.all()
    serializer_class = RiskIncidentTypeRepresentationSerializer
    search_fields = ("name",)


class RiskIncidentRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = RiskIncident.objects.all()
    serializer_class = RiskIncidentRepresentationSerializer
    search_fields = (
        "rule__name",
        "breached_object_repr",
    )


class CheckedObjectIncidentRelationshipRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = CheckedObjectIncidentRelationship.objects.all()
    serializer_class = CheckedObjectIncidentRelationshipRepresentationSerializer


class RiskIncidentModelViewSet(viewsets.ModelViewSet):
    permission_classes = [RulePermission]
    display_config_class = RiskIncidentDisplayConfig
    endpoint_config_class = RiskIncidentEndpointConfig
    button_config_class = RiskIncidentButtonConfig

    ordering_fields = [
        "status",
        "date_range",
        "rule__name",
        "object_repr",
        "severity__name",
        "resolved_by__computed_str",
        "comment",
    ]
    ordering = ["-date_range", "-severity__severity_order", "id"]
    serializer_class = RiskIncidentModelSerializer
    queryset = RiskIncident.objects.select_related(
        "resolved_by",
        "breached_content_type",
        "rule",
        "severity",
    ).prefetch_related("checked_object_relationships")
    filterset_class = RiskIncidentFilterSet

    @action(detail=False, methods=["PATCH"])
    def resolveallincidents(self, request, pk=None):
        resolve_status = request.POST.get("resolve_status", RiskIncident.Status.RESOLVED.name)
        resolve_all_incidents_as_task.delay(
            request.user.id,
            request.POST.get("comment", ""),
            resolve_status == RiskIncident.Status.RESOLVED.name,
            rule_id=request.GET.get("rule_id", None),
        )
        return HttpResponse("All opened incidents are resolved", status=status.HTTP_200_OK)

    @cached_property
    def rule_content_type(self):
        return ContentType.objects.get_for_model(RiskRule)

    @cached_property
    def view_rule_permission(self):
        return Permission.objects.get(codename="view_riskrule", content_type__app_label="wbcompliance")

    def get_queryset(self):
        """
        We protect the queryset to allow only user to see their respective incidents
        """
        queryset = super().get_queryset()
        if not self.request.user.has_perm("administrate_riskrule"):
            queryset = queryset.annotate(
                rule_pk=Cast(
                    "rule_id", output_field=CharField()
                ),  # We need to cast the rule id into a charfield otherwise postgres complains with the operator
                have_view_permission_on_rule=Exists(
                    UserObjectPermission.objects.filter(
                        permission=self.view_rule_permission,
                        content_type=self.rule_content_type,
                        object_pk=OuterRef("rule_pk"),
                        user=self.request.user,
                    )
                ),
            ).filter(have_view_permission_on_rule=True)
        queryset = queryset.annotate(
            status_icon=Case(
                When(status=RiskIncident.Status.OPEN.value, then=Value(WBIcon.VIEW.icon)), default=Value(None)
            ),
            has_subincidents=Exists(CheckedObjectIncidentRelationship.objects.filter(incident=OuterRef("pk"))),
            _group_key=Case(When(has_subincidents=True, then=F("id")), default=None, output_field=IntegerField()),
            checked_date=ExpressionWrapper(F("date_range__endswith") - 1, output_field=DateField()),
            object_repr=Concat(Value("Breached: "), F("breached_object_repr")),
            threshold_repr=Subquery(
                RuleThreshold.objects.filter(severity=OuterRef("severity"), rule=OuterRef("rule")).values(
                    "computed_str"
                )[:1]
            ),
        )
        return queryset

    def add_messages(self, request, instance: RiskIncident | None = None, **kwargs):
        if instance:
            if instance.status == RiskIncident.Status.OPEN and instance.last_ignored_date:
                warning(
                    request,
                    f"This incident was automatically reopened after being ignored from {instance.last_ignored_date:%Y-%m-%d} to {instance.get_ignore_until_date():%Y-%m-%d}",
                )


class RiskIncidentRiskRuleModelViewSet(RiskIncidentModelViewSet):
    permission_classes = [RulePermission]

    def get_queryset(self):
        return super().get_queryset().filter(rule=self.kwargs["rule_id"])


class CheckedObjectIncidentRelationshipModelViewSet(viewsets.ModelViewSet):
    permission_classes = [RulePermission]
    ordering = ["-rule_check__evaluation_date"]
    ordering_fields = ["status", "object_repr", "checked_date"]
    serializer_class = CheckedObjectIncidentRelationshipModelSerializer
    queryset = CheckedObjectIncidentRelationship.objects.select_related(
        "resolved_by",
        "incident",
        "rule_check",
        "severity",
    ).annotate(
        object_repr=F("rule_check__checked_object_repr"),
        checked_date=F("rule_check__evaluation_date"),
        date_range=Cast(
            Concat(Value("["), F("checked_date"), Value(","), F("checked_date"), Value("]"), output_field=CharField()),
            output_field=DateRangeField(),
        ),
        threshold_repr=Subquery(
            RuleThreshold.objects.filter(severity=OuterRef("severity"), rule=OuterRef("incident__rule")).values(
                "computed_str"
            )[:1]
        ),
    )
    display_config_class = CheckedObjectIncidentRelationshipRiskRuleDisplayConfig
    endpoint_config_class = CheckedObjectIncidentRelationshipEndpointConfig
    filterset_class = CheckedObjectIncidentRelationshipFilterSet


class CheckedObjectIncidentRelationshipRiskRuleModelViewSet(CheckedObjectIncidentRelationshipModelViewSet):
    endpoint_config_class = CheckedObjectIncidentRelationshipRiskRuleEndpointConfig

    def get_queryset(self):
        return super().get_queryset().filter(incident=self.kwargs["incident_id"])



from django.utils.translation import gettext as _
from wbcore.metadata.configs.titles import TitleViewConfig


class RiskManagementIncidentTableTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return _("Risk Managements incident table")



# Generated by Django 5.0.12 on 2025-11-28 09:18

from django.db import migrations, models
import django.db.models.deletion
from django.db.models import Exists, OuterRef


def handle_data(apps, schema_editor):
    CheckedObjectIncidentRelationship = apps.get_model('wbcompliance', 'CheckedObjectIncidentRelationship')
    RiskIncident = apps.get_model('wbcompliance', 'RiskIncident')
    CheckedObjectIncidentRelationship.objects.filter(incident__isnull=True).delete()
    RiskIncident.objects.filter(checked_object_relationships__isnull=True).delete()
    RiskIncident.objects.annotate(
        is_precheck=Exists(CheckedObjectIncidentRelationship.objects.filter(incident=OuterRef("id"), rule_check__activator_id__isnull=False))
    ).filter(is_precheck=True).update(precheck=True)

class Migration(migrations.Migration):

    dependencies = [
        ('wbcompliance', '0021_riskcheck_passive_check'),
    ]

    operations = [
        migrations.AddField(
            model_name='riskincident',
            name='precheck',
            field=models.BooleanField(default=False, help_text='If true, this incident was created during a precheck evaluation', verbose_name='Precheck Incident'),
        ),
        migrations.RunSQL(sql="SET CONSTRAINTS ALL IMMEDIATE;"),
        migrations.RunPython(handle_data),
        migrations.RunSQL(sql="SET CONSTRAINTS ALL DEFERRED;"),
        migrations.AlterField(
            model_name='checkedobjectincidentrelationship',
            name='incident',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE,
                                    related_name='checked_object_relationships', to='wbcompliance.riskincident'),
            preserve_default=False,
        ),
    ]



# Generated by Django 5.0.13 on 2025-03-13 07:51

import django.db.models.deletion
from django.db import migrations, models
from tqdm import tqdm
def migrate_risk_check(apps, schema_editor):

    RiskCheck = apps.get_model("wbcompliance", "RiskCheck")
    objs = []
    qs = RiskCheck.objects.all()
    for check in tqdm(qs, total=qs.count()):
        check.rule_id = check.rule_checked_object_relationship.rule.id
        check.checked_object_content_type_id = check.rule_checked_object_relationship.checked_object_content_type.id
        check.checked_object_id = check.rule_checked_object_relationship.checked_object_id
        check.save()
    #     objs.append(check)
    #     if len(objs) > 1000:
    #         RiskCheck.objects.bulk_update(objs, ["rule_id", "checked_object_content_type_id", "checked_object_id"])
    #         objs = []
    # RiskCheck.objects.bulk_update(objs, ["rule_id", "checked_object_content_type_id", "checked_object_id"])

class Migration(migrations.Migration):

    dependencies = [
        ('wbcompliance', '0019_rulegroup_riskrule_activation_date_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='riskcheck',
            name='rule',
            field=models.ForeignKey(null=True, blank=True, on_delete=django.db.models.deletion.CASCADE, related_name='checks',
                                    to='wbcompliance.riskrule'),
            preserve_default=False,
        ),
        migrations.RenameField(
            model_name='riskcheck',
            old_name='trigger_content_type',
            new_name='checked_object_content_type',
        ),
        migrations.RenameField(
            model_name='riskcheck',
            old_name='trigger_id',
            new_name='checked_object_id',
        ),
        migrations.AddField(
            model_name='riskcheck',
            name='checked_object_repr',
            field=models.CharField(blank=True, max_length=256, null=True),
        ),
        migrations.RunSQL(sql="SET CONSTRAINTS ALL IMMEDIATE;"),
        migrations.RunPython(migrate_risk_check),
        migrations.RunSQL(sql="SET CONSTRAINTS ALL DEFERRED;"),
        migrations.RemoveField(
            model_name='riskcheck',
            name='rule_checked_object_relationship',
        ),
        migrations.AlterField(
            model_name='riskcheck',
            name='rule',
            field=models.ForeignKey( on_delete=django.db.models.deletion.CASCADE, related_name='checks', to='wbcompliance.riskrule'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='riskcheck',
            name='checked_object_content_type',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE,
                                    related_name='triggered_checks', to='contenttypes.contenttype'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='riskcheck',
            name='checked_object_id',
            field=models.PositiveIntegerField(default=None),
            preserve_default=False,
        ),
    ]
